from dataclasses import dataclass, field

from bioexperiment_suite.experiment import Experiment

from .device_widgets import PumpWidget, SpectrophotometerWidget


@dataclass
class Store:
    experiment: Experiment = field(default_factory=Experiment)

    pump_widgets: list[PumpWidget] = field(default_factory=list)

    @property
    def pump_names(self) -> list[str]:
        return [widget.title.get() for widget in self.pump_widgets]

    def get_pump_by_name(self, pump_name: str):
        for widget in self.pump_widgets:
            if widget.title.get() == pump_name:
                return widget.device

    spectrophotometer_widgets: list[SpectrophotometerWidget] = field(default_factory=list)

    @property
    def spectrophotometer_names(self) -> list[str]:
        return [widget.title.get() for widget in self.spectrophotometer_widgets]

    def get_spectrophotometer_by_name(self, spectrophotometer_name: str):
        for widget in self.spectrophotometer_widgets:
            if widget.title.get() == spectrophotometer_name:
                return widget.device



from abc import ABC
from datetime import datetime
from tkinter import DoubleVar, StringVar, simpledialog
from typing import Callable

import ttkbootstrap as ttk
from bioexperiment_suite.interfaces import Pump, SerialConnection, Spectrophotometer
from ttkbootstrap import constants as c


class DeviceWidget(ttk.Labelframe, ABC):
    instance_count = 0  # Class variable to keep track of the number of instances
    DEVICE_TYPE = ""
    FRAME_PADDING = 5
    PADX = 5
    PADY = 5

    def __init__(self, parent: ttk.Frame, device: SerialConnection):
        super().__init__(parent, padding=self.FRAME_PADDING, bootstyle=c.PRIMARY, text=self.DEVICE_TYPE)
        self.device = device
        self.title = StringVar(value=f"{self.DEVICE_TYPE} {self.instance_count}")

    def crate_title_frame(self) -> ttk.Frame:
        frame = ttk.Frame(self, padding=self.FRAME_PADDING)

        label = ttk.Label(frame, textvariable=self.title, font=("Helvetica", 16, "bold"))
        label.pack(side=c.LEFT, padx=self.PADX, pady=self.PADY)

        rename_button = ttk.Button(frame, text="Rename", command=self.rename, bootstyle=c.SECONDARY)
        rename_button.pack(side=c.RIGHT, padx=self.PADX, pady=self.PADY)

        return frame

    def rename(self):
        new_name = simpledialog.askstring(
            f"Rename {self.DEVICE_TYPE}",
            "Enter new name:",
            initialvalue=self.title.get(),
        )
        if new_name:
            self.title.set(new_name)

    def create_info_section(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.INFO, text="Info", padding=self.FRAME_PADDING)
        frame.columnconfigure(0, weight=1, minsize=110)
        frame.columnconfigure(1, weight=1, minsize=40)

        port_name_label = ttk.Label(frame, text="Port:")
        port_name_label.grid(row=0, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        port_name_info = ttk.Label(frame, text=self.device.port)
        port_name_info.grid(row=0, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        baudrate_label = ttk.Label(frame, text="Baudrate:")
        baudrate_label.grid(row=1, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        baudrate_info = ttk.Label(frame, text=self.device.baudrate)
        baudrate_info.grid(row=1, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        return frame


class PumpWidget(DeviceWidget):
    DEVICE_TYPE = "Pump"

    def __init__(self, parent: ttk.Frame, pump: Pump):
        PumpWidget.instance_count += 1
        super().__init__(parent, pump)

        self.pump = pump
        self.flow_rate = DoubleVar(value=3.0)
        self.volume = DoubleVar(value=0.0)
        self.direction = StringVar(value="right")

        self.set_flow_rate()
        self.create_widgets()

    def create_flow_rate_control(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.PRIMARY, text="Flow rate control", padding=self.FRAME_PADDING)
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        label = ttk.Label(frame, text="Flow rate (mL/min):")
        label.grid(row=0, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        entry = ttk.Entry(frame, textvariable=self.flow_rate, bootstyle=c.PRIMARY, width=label.winfo_width())
        entry.grid(row=0, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        set_button = ttk.Button(frame, text="Set", command=self.set_flow_rate, bootstyle=c.PRIMARY)
        set_button.grid(row=1, column=0, sticky="ew", columnspan=2, padx=self.PADX, pady=self.PADY)

        return frame

    def create_continuous_rotation_control(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(
            self, bootstyle=c.PRIMARY, text="Continuous rotation control", padding=self.FRAME_PADDING
        )
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(2, weight=1)

        rotate_right_button = ttk.Button(
            frame,
            text="Rotate right",
            command=lambda: self.pump.start_continuous_rotation(direction="right"),
            bootstyle=c.SUCCESS,
        )
        rotate_right_button.grid(row=0, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        stop_button = ttk.Button(frame, text="Stop", bootstyle=c.DANGER, command=self.pump.stop_continuous_rotation)
        stop_button.grid(row=0, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        rotate_left_button = ttk.Button(
            frame,
            text="Rotate Left",
            command=lambda: self.pump.start_continuous_rotation(direction="left"),
            bootstyle=c.SUCCESS,
        )
        rotate_left_button.grid(row=0, column=2, sticky="ew", padx=self.PADX, pady=self.PADY)

        return frame

    def create_volume_pouring_control(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.PRIMARY, text="Volume pouring control", padding=self.FRAME_PADDING)
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        volume_label = ttk.Label(frame, text="Volume (mL):")
        volume_label.grid(row=0, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        volume_entry = ttk.Entry(frame, textvariable=self.volume, bootstyle=c.PRIMARY, width=volume_label.winfo_width())
        volume_entry.grid(row=0, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        direction_frame = ttk.Frame(frame, padding=0)
        direction_frame.grid(row=1, column=0, columnspan=2, sticky="ew")
        direction_frame.columnconfigure(0, weight=1)
        direction_frame.columnconfigure(1, weight=1)
        right_radio = ttk.Radiobutton(
            direction_frame, text="Right", variable=self.direction, value="right", bootstyle=("toolbutton", c.PRIMARY)
        )
        right_radio.grid(row=0, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        left_radio = ttk.Radiobutton(
            direction_frame, text="Left", variable=self.direction, value="left", bootstyle=("toolbutton", c.PRIMARY)
        )
        left_radio.grid(row=0, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        pump_button = ttk.Button(
            frame,
            text="Pump",
            command=lambda: self.pump.pour_in_volume(self.volume.get(), direction=self.direction.get()),
            bootstyle=c.SUCCESS,
        )
        pump_button.grid(row=2, column=0, columnspan=2, sticky="ew", padx=self.PADX, pady=self.PADY)

        return frame

    def create_widgets(self):
        title_frame = self.crate_title_frame()
        title_frame.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        info_section = self.create_info_section()
        info_section.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        flow_rate_control = self.create_flow_rate_control()
        flow_rate_control.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        continuous_rotation_control = self.create_continuous_rotation_control()
        continuous_rotation_control.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        volume_pouring_control = self.create_volume_pouring_control()
        volume_pouring_control.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

    def set_flow_rate(self):
        flow_rate = self.flow_rate.get()
        self.pump.set_default_flow_rate(flow_rate)
        print(f"Flow rate set to {flow_rate} mL/min")


class SpectrophotometerWidget(DeviceWidget):
    DEVICE_TYPE = "Spectrophotometer"

    def __init__(self, parent: ttk.Frame, spectrophotometer: Spectrophotometer):
        SpectrophotometerWidget.instance_count += 1
        super().__init__(parent, spectrophotometer)

        self.spectrophotometer = spectrophotometer

        self.temperature = DoubleVar()
        self.temperature_updated = StringVar(value="-")

        self.optical_density = DoubleVar()
        self.optical_density_updated = StringVar(value="-")

        self.create_widgets()

    def create_measurement_display(
        self, title: str, value_var: DoubleVar, updated_var: StringVar, update_function: Callable
    ) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.PRIMARY, text=title, padding=self.FRAME_PADDING)
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        label = ttk.Label(frame, text=f"{title}:")
        label.grid(row=0, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        value = ttk.Label(frame, textvariable=value_var)
        value.grid(row=0, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        time_label = ttk.Label(frame, text="Last updated:")
        time_label.grid(row=1, column=0, sticky="ew", padx=self.PADX, pady=self.PADY)

        time_value = ttk.Label(frame, textvariable=updated_var)
        time_value.grid(row=1, column=1, sticky="ew", padx=self.PADX, pady=self.PADY)

        measure_button = ttk.Button(frame, text="Measure", command=update_function, bootstyle=c.PRIMARY)
        measure_button.grid(row=2, column=0, columnspan=2, sticky="ew", padx=self.PADX, pady=self.PADY)

        return frame

    def create_widgets(self):
        title_frame = self.crate_title_frame()
        title_frame.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        info_section = self.create_info_section()
        info_section.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        temperature_display = self.create_measurement_display(
            "Temperature, °C", self.temperature, self.temperature_updated, self.measure_temperature
        )
        temperature_display.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        optical_density_display = self.create_measurement_display(
            "Optical density", self.optical_density, self.optical_density_updated, self.measure_optical_density
        )
        optical_density_display.pack(fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

    def measure_temperature(self):
        temperature = self.spectrophotometer.get_temperature()
        self.temperature_updated.set(datetime.now().strftime("%H:%M:%S"))
        self.temperature.set(round(temperature, 2))

    def measure_optical_density(self):
        optical_density = self.spectrophotometer.measure_optical_density()
        self.optical_density_updated.set(datetime.now().strftime("%H:%M:%S"))
        self.optical_density.set(round(optical_density, 5))



from typing import Callable, SupportsIndex, Optional, TypeAlias
from statistics import median

RelationFunction: TypeAlias = Callable[[float], bool]


class Relation:
    """Class to define common relations for conditional actions."""

    @staticmethod
    def EQUALS_TO(value: float) -> RelationFunction:
        """Returns a relation checking if a metric equals the given value."""
        return lambda x: x == value

    @staticmethod
    def GREATER_THAN(value: float) -> RelationFunction:
        """Returns a relation checking if a metric is greater than the given value."""
        return lambda x: x > value

    @staticmethod
    def LESS_THAN(value: float) -> RelationFunction:
        """Returns a relation checking if a metric is less than the given value."""
        return lambda x: x < value

    @staticmethod
    def GREATER_THAN_OR_EQUALS_TO(value: float) -> RelationFunction:
        """Returns a relation checking if a metric is greater than or equal to the given value."""
        return lambda x: x >= value

    @staticmethod
    def LESS_THAN_OR_EQUALS_TO(value: float) -> RelationFunction:
        """Returns a relation checking if a metric is less than or equal to the given value."""
        return lambda x: x <= value

    @staticmethod
    def NOT_EQUALS_TO(value: float) -> RelationFunction:
        """Returns a relation checking if a metric does not equal the given value."""
        return lambda x: x != value


class Statistic:
    """A class that provides common statistical functions."""

    @staticmethod
    def LAST() -> Callable[[SupportsIndex], float]:
        return lambda measurement: measurement[-1]

    @staticmethod
    def COUNT() -> Callable[[SupportsIndex], int]:
        return len

    @staticmethod
    def SUM(window_size: Optional[int] = None) -> Callable[[SupportsIndex], float]:
        return lambda measurement: sum(measurement[-window_size:]) if window_size else sum(measurement)

    @staticmethod
    def MEAN(window_size: Optional[int] = None) -> Callable[[SupportsIndex], float]:
        return lambda measurement: (
            sum(measurement[-window_size:]) / window_size if window_size else sum(measurement) / len(measurement)
        )

    @staticmethod
    def MEDIAN(window_size: Optional[int] = None) -> Callable[[SupportsIndex], float]:
        return lambda measurement: median(measurement[-window_size:]) if window_size else median(measurement)

    @staticmethod
    def MAX(window_size: Optional[int] = None) -> Callable[[SupportsIndex], float]:
        return lambda measurement: max(measurement[-window_size:]) if window_size else max(measurement)

    @staticmethod
    def MIN(window_size: Optional[int] = None) -> Callable[[SupportsIndex], float]:
        return lambda measurement: min(measurement[-window_size:]) if window_size else min(measurement)



from typing import Type

import ttkbootstrap as ttk
from ttkbootstrap import constants as c

from .devices_tab import ConnectedDevicesWidget
from .experiment_tab import ExperimentWidget
from .store import Store


class MainWindow(ttk.Frame):
    def __init__(self, master: ttk.Window):
        super().__init__(master)
        self.master = master
        self.master.title("Bioexperiment Suite")
        self.pack(fill=c.BOTH, expand=c.YES)

        self.store = Store()

        self.notebook = ttk.Notebook(self)
        self.notebook.pack(fill=c.BOTH, expand=c.YES)

        self.add_tab(ConnectedDevicesWidget, "Devices")
        self.add_tab(ExperimentWidget, "Experiment")

    def add_tab(self, tab_constructor: Type[ttk.Frame], title: str):
        tab = tab_constructor(self.notebook, self.store)
        tab.pack(fill=c.BOTH, expand=c.YES)
        self.notebook.add(tab, text=title)


def main():
    root = ttk.Window(minsize=(1000, 600), themename="litera")
    MainWindow(root)
    root.mainloop()


if __name__ == "__main__":
    main()



from pathlib import Path
from tkinter import DoubleVar, Event, IntVar, StringVar, filedialog, messagebox

import ttkbootstrap as ttk
from bioexperiment_suite.interfaces import Pump, Spectrophotometer
from bioexperiment_suite.loader import logger
from ttkbootstrap import constants as c

from .store import Store


class ExperimentWidget(ttk.Frame):
    FRAME_PADDING = 5
    PADX = 5
    PADY = 5

    def __init__(self, master: ttk.Frame, store: Store):
        super().__init__(master)

        self.store = store

        self.infuse_pump: Pump | None = None
        self.pour_out_pump: Pump | None = None
        self.spectrophotometer: Spectrophotometer | None = None

        self.output_directory_path = StringVar(value="Didn't selected (No CSV output)")
        self.experiment_duration_hours = IntVar(value=24)
        self.solution_refresh_interval_minutes = IntVar(value=60)
        self.measurement_interval_minutes = IntVar(value=5)
        self.poured_out_volume_ml = DoubleVar(value=2.0)
        self.infused_volume_ml = DoubleVar(value=1.0)
        self.flow_rate_ml_per_minute = DoubleVar(value=3.0)

        self.is_experiment_setup = False
        self.is_experiment_running = False

        self.create_widgets()

    def ask_for_output_directory(self):
        self.output_directory_path.set(filedialog.askdirectory(initialdir=".", title="Select results output directory"))
        if not self.output_directory_path.get():
            self.output_directory_path.set("Didn't selected (No CSV output)")
            return
        output_directory_path = Path(self.output_directory_path.get())
        self.store.experiment.specify_output_dir(output_directory_path)

    def create_output_directory_widget(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.PRIMARY, text="CSV Output", padding=self.FRAME_PADDING)
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        output_dir_label = ttk.Label(frame, text="Output Directory:")
        output_dir_label.grid(row=0, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        output_dir_button = ttk.Button(frame, text="Select", command=self.ask_for_output_directory)
        output_dir_button.grid(row=0, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        output_dir_entry = ttk.Label(frame, textvariable=self.output_directory_path, style="info.TLabel")
        output_dir_entry.grid(row=1, column=0, columnspan=2, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        return frame

    def create_experiment_setup_widget(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.PRIMARY, text="Experiment Parameters", padding=self.FRAME_PADDING)
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        duration_label = ttk.Label(frame, text="Total duration (hours):")
        duration_label.grid(row=0, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        duration_entry = ttk.Entry(frame, textvariable=self.experiment_duration_hours)
        duration_entry.grid(row=0, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        refresh_interval_label = ttk.Label(frame, text="Solution refresh interval (minutes):")
        refresh_interval_label.grid(row=1, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        refresh_interval_entry = ttk.Entry(frame, textvariable=self.solution_refresh_interval_minutes)
        refresh_interval_entry.grid(row=1, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        measurement_interval_label = ttk.Label(frame, text="Measurement interval (minutes):")
        measurement_interval_label.grid(row=2, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        measurement_interval_entry = ttk.Entry(frame, textvariable=self.measurement_interval_minutes)
        measurement_interval_entry.grid(row=2, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        poured_out_volume_label = ttk.Label(frame, text="Poured out volume (mL):")
        poured_out_volume_label.grid(row=3, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        poured_out_volume_entry = ttk.Entry(frame, textvariable=self.poured_out_volume_ml)
        poured_out_volume_entry.grid(row=3, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        infused_volume_label = ttk.Label(frame, text="Infused volume (mL):")
        infused_volume_label.grid(row=4, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        infused_volume_entry = ttk.Entry(frame, textvariable=self.infused_volume_ml)
        infused_volume_entry.grid(row=4, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        flow_rate_label = ttk.Label(frame, text="Flow rate (mL/min):")
        flow_rate_label.grid(row=5, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        flow_rate_entry = ttk.Entry(frame, textvariable=self.flow_rate_ml_per_minute)
        flow_rate_entry.grid(row=5, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        return frame

    def create_devices_choice_widget(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.PRIMARY, text="Devices", padding=self.FRAME_PADDING)
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        infused_pump_label = ttk.Label(frame, text="Infuse Pump:")
        infused_pump_label.grid(row=0, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        poured_out_pump_label = ttk.Label(frame, text="Pour Out Pump:")
        poured_out_pump_label.grid(row=1, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        spectrophotometer_label = ttk.Label(frame, text="Spectrophotometer:")
        spectrophotometer_label.grid(row=2, column=0, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        def handle_infuse_pump_choice(event: Event) -> None:
            self.infuse_pump = self.store.get_pump_by_name(event.widget.get())

        def handle_pour_out_pump_choice(event: Event) -> None:
            self.pour_out_pump = self.store.get_pump_by_name(event.widget.get())

        def handle_spectrophotometer_choice(event: Event) -> None:
            self.spectrophotometer = self.store.get_spectrophotometer_by_name(event.widget.get())

        def render_choices():
            infuse_pump_choice = ttk.Combobox(
                frame,
                values=self.store.pump_names,
                state="readonly",
            )
            infuse_pump_choice.grid(row=0, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)
            infuse_pump_choice.bind("<<ComboboxSelected>>", handle_infuse_pump_choice)

            pour_out_pump_choice = ttk.Combobox(
                frame,
                values=self.store.pump_names,
                state="readonly",
            )
            pour_out_pump_choice.grid(row=1, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)
            pour_out_pump_choice.bind("<<ComboboxSelected>>", handle_pour_out_pump_choice)

            spectrophotometer_choice = ttk.Combobox(
                frame,
                values=self.store.spectrophotometer_names,
                state="readonly",
            )
            spectrophotometer_choice.grid(row=2, column=1, sticky=c.EW, padx=self.PADX, pady=self.PADY)
            spectrophotometer_choice.bind("<<ComboboxSelected>>", handle_spectrophotometer_choice)

        render_choices()

        update_button = ttk.Button(frame, text="Update", command=render_choices)
        update_button.grid(row=3, column=0, columnspan=2, sticky=c.EW, padx=self.PADX, pady=self.PADY)

        return frame

    def setup_experiment(self):
        MEASUREMENT_WAIT_TIME_SECONDS = 60
        n_solution_refreshes = self.experiment_duration_hours.get() * 60 // self.solution_refresh_interval_minutes.get()
        n_measurements_per_solution_refresh = (
            self.solution_refresh_interval_minutes.get() // self.measurement_interval_minutes.get()
        )

        for _ in range(n_solution_refreshes):
            for _ in range(n_measurements_per_solution_refresh):
                self.store.experiment.add_measurement(
                    self.spectrophotometer.get_temperature,  # type: ignore
                    measurement_name="Temperature (C)",
                )
                self.store.experiment.add_measurement(
                    self.spectrophotometer.measure_optical_density,  # type: ignore
                    measurement_name="Optical density",
                )
                self.store.experiment.add_wait(self.measurement_interval_minutes.get() * 60)

            self.store.experiment.actions.pop()
            self.store.experiment.add_wait(MEASUREMENT_WAIT_TIME_SECONDS)
            self.store.experiment.add_action(
                self.pour_out_pump.pour_in_volume,  # type: ignore
                volume=self.poured_out_volume_ml.get(),
                flow_rate=self.flow_rate_ml_per_minute.get(),
                direction="left",
            )
            self.store.experiment.add_action(
                self.infuse_pump.pour_in_volume,  # type: ignore
                volume=self.infused_volume_ml.get(),
                flow_rate=self.flow_rate_ml_per_minute.get(),
                direction="right",
            )
            self.store.experiment.add_wait(self.measurement_interval_minutes.get() * 60 - MEASUREMENT_WAIT_TIME_SECONDS)

    def apply_experiment_settings(self):
        if not self.infuse_pump or not self.pour_out_pump or not self.spectrophotometer:
            messagebox.showerror(title="Error", message="Please select all devices")
            return

        try:
            assert self.experiment_duration_hours.get() > 0
            assert self.solution_refresh_interval_minutes.get() > 0
            assert self.measurement_interval_minutes.get() > 0
        except AssertionError:
            messagebox.showerror(title="Error", message="Please enter positive values for durations and intervals")
            return

        self.store.experiment.reset_experiment()
        self.setup_experiment()
        self.is_experiment_setup = True
        logger.info("Experiment is ready to start")

    def start_experiment(self):
        if not self.is_experiment_setup:
            messagebox.showerror(title="Error", message="Please apply experiment settings first")
            return

        self.store.experiment.start()
        self.is_experiment_running = True

    def stop_experiment(self):
        if not self.is_experiment_running:
            messagebox.showerror(title="Error", message="Experiment is not running")
            return

        self.store.experiment.stop()
        self.is_experiment_running = False

    def create_experiment_controls_widget(self) -> ttk.Labelframe:
        frame = ttk.Labelframe(self, bootstyle=c.PRIMARY, text="Experiment Controls", padding=self.FRAME_PADDING)

        apply_settings_button = ttk.Button(
            frame, text="Apply Experiment Settings", command=self.apply_experiment_settings, bootstyle=c.PRIMARY
        )
        apply_settings_button.pack(fill=c.X, padx=self.PADX, pady=self.PADY)

        start_experiment_button = ttk.Button(
            frame, text="Start Experiment", command=self.start_experiment, bootstyle=c.SUCCESS
        )
        start_experiment_button.pack(fill=c.X, padx=self.PADX, pady=self.PADY)

        stop_experiment_button = ttk.Button(
            frame, text="Stop Experiment", command=self.stop_experiment, bootstyle=c.DANGER
        )
        stop_experiment_button.pack(fill=c.X, padx=self.PADX, pady=self.PADY)

        return frame

    def create_widgets(self):
        output_directory_widget = self.create_output_directory_widget()
        output_directory_widget.grid(row=1, column=0, sticky=c.NSEW, padx=self.PADX, pady=self.PADY)

        devices_choice_widget = self.create_devices_choice_widget()
        devices_choice_widget.grid(row=0, column=0, sticky=c.NSEW, padx=self.PADX, pady=self.PADY)

        experiment_setup_widget = self.create_experiment_setup_widget()
        experiment_setup_widget.grid(row=0, column=1, sticky=c.NSEW, padx=self.PADX, pady=self.PADY)

        experiment_controls_widget = self.create_experiment_controls_widget()
        experiment_controls_widget.grid(row=1, column=1, sticky=c.NSEW, padx=self.PADX, pady=self.PADY)



from random import random
from time import sleep

from bioexperiment_suite.loader import device_interfaces, logger
from bioexperiment_suite.settings import settings

from .serial_connection import SerialConnection


class Spectrophotometer(SerialConnection):
    """Class to handle communication with a spectrophotometer connected to a serial port."""

    def __init__(self, port: str, baudrate: int = 9600, timeout_sec: int | float = 1.0):
        """Initializes the spectrophotometer object.

        :param port: The serial port to connect to
        :param baudrate: The baudrate of the serial connection. Defaults to 9600
        :param timeout_sec: The timeout of the serial connection to respond in seconds. Defaults to 1.0
        """
        self.interface = device_interfaces.spectrophotometer
        super(Spectrophotometer, self).__init__(port, baudrate, timeout_sec)

    def get_temperature(self) -> float:
        """Gets the temperature of the spectrophotometer

        :returns: The temperature in degrees Celsius
        """

        if settings.EMULATE_DEVICES:
            logger.debug("Getting FAKE temperature")
            temperature = random() * 10 + 20
            logger.debug(f"FAKE temperature: {temperature:.2f}")
            return temperature

        logger.debug("Getting temperature")
        temperature_response = self.communicate_with_serial_port(
            self.interface.commands.get_temperature.request,
            self.interface.commands.get_temperature.response_len,
        )
        logger.debug(f"Temperature response: {list(temperature_response)}")
        integer, fractional = temperature_response[2:]
        temperature = integer + (fractional / 100)
        logger.debug(f"Temperature: {temperature:.2f}")
        return temperature

    def _send_start_measurement_command(self):
        """Sends the command to start the measurement."""
        self.write_to_serial_port(self.interface.commands.start_measurement.request)
        logger.debug("Start measurement command sent")

    def _get_optical_density(self) -> float | None:
        """Gets the optical density of the sample.

        :returns: The optical density of the sample
        """

        if settings.EMULATE_DEVICES:
            logger.debug("Getting FAKE optical density")
            optical_density = random()
            logger.debug(f"Fake optical density: {optical_density:.5f}")
            return optical_density

        optical_density_response = self.communicate_with_serial_port(
            self.interface.commands.get_measurement_result.request,
            self.interface.commands.get_measurement_result.response_len,
        )
        logger.debug(f"Optical density response: {list(optical_density_response)}")
        if not optical_density_response:
            return None
        integer, fractional = optical_density_response[2:]
        optical_density = integer + (fractional / 100)
        logger.debug(f"Optical density: {optical_density:.5f}")
        return optical_density

    def measure_optical_density(self) -> float:
        """Measures the optical density of the sample.

        :returns: The optical density of the sample
        """
        logger.debug("Measuring optical density")
        self._send_start_measurement_command()
        logger.debug("Optical density not ready yet, waiting...")
        sleep(3 if not settings.EMULATE_DEVICES else 1)
        optical_density = self._get_optical_density()
        if optical_density is None:
            logger.error("Optical density could not be measured")
            raise Exception("Optical density could not be measured")

        return optical_density



import glob
import sys

import serial

from bioexperiment_suite.loader import logger
from bioexperiment_suite.settings import settings


def get_serial_ports() -> list[str]:
    """Lists serial port names on the system.

    :returns: A list of the serial ports available on the system

    :raises EnvironmentError: On unsupported or unknown platforms
    """
    if sys.platform.startswith("win"):
        logger.info("Windows platform detected")
        ports = [f"COM{i + 1}" for i in range(256)]
    elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
        logger.info("Linux platform detected")
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob("/dev/tty[A-Za-z]*")
    elif sys.platform.startswith("darwin"):
        logger.info("MacOS platform detected")
        ports = glob.glob("/dev/tty.*")
    else:
        logger.error(f"Unsupported platform: {sys.platform}")
        raise EnvironmentError("Unsupported platform")

    result = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass

    return result


if settings.EMULATE_DEVICES:
    # Overwrite the function if EMULATE_DEVICES is True
    def get_serial_ports(n_pumps_n_spectrophotometers: tuple[int, int]) -> list[str]:
        """Lists serial port names on the system.

        :param n_pumps_n_spectrophotometers: A tuple containing the number of pumps and spectrophotometers to find

        :returns: A list of the serial ports available on the system

        :raises EnvironmentError: On unsupported or unknown platforms
        """
        pumps_port_n, spectrophotometers_port_n = n_pumps_n_spectrophotometers
        pumps_port_numbers = [i * 2 for i in range(pumps_port_n)]
        spectrophotometers_port_numbers = [i * 2 + 1 for i in range(spectrophotometers_port_n)]
        result = [f"COM{i}" for i in sorted(pumps_port_numbers + spectrophotometers_port_numbers)]
        logger.debug(f"Fake serial ports found: {result}")
        return result



from __future__ import annotations

from functools import wraps
from time import sleep
from typing import Callable

import serial

from bioexperiment_suite.loader import logger
from bioexperiment_suite.settings import settings


class SerialConnection:
    """Class to handle serial communication with a device connected to a serial port."""

    def __init__(self, port: str, baudrate: int = 9600, timeout_sec: float = 1.0):
        """Initializes the serial connection object.

        :param port: The serial port to connect to
        :param baudrate: The baudrate of the serial connection. Defaults to 9600
        :param timeout_sec: The timeout of the serial connection to respond in seconds. Defaults to 1.0

        :raises serial.SerialException: If the serial connection cannot be established
        """
        self.port = port
        self.baudrate = baudrate
        self.timeout_sec = timeout_sec
        self._create_serial_connection()

    def _create_serial_connection(self) -> None:
        """Creates a serial connection with the specified parameters."""

        if settings.EMULATE_DEVICES:
            logger.info(f"FAKE serial connection established with {self.port}")
            sleep(1)
            return

        self.serial = serial.Serial(self.port, self.baudrate, timeout=self.timeout_sec)
        logger.info(f"Serial connection established with {self.port}")
        sleep(3)

    @staticmethod
    def _restore_connection(method: Callable) -> Callable:
        """Decorator to restore the serial connection if it is lost during communication.

        :param method: The method to decorate

        :returns: The decorated method
        """

        @wraps(method)
        def wrapper(self: SerialConnection, *args, **kwargs):
            try:
                return method(self, *args, **kwargs)
            except serial.SerialException:
                logger.warning(f"Serial connection lost on port {self.port}. Restoring connection...")
                self._create_serial_connection()
                return method(self, *args, **kwargs)

        return wrapper

    @_restore_connection
    def write_to_serial_port(self, data_to_send: list[int]) -> None:
        """Writes data to the serial port.

        :param data_to_send: The data to send to the serial port
        """
        if settings.EMULATE_DEVICES:
            logger.debug(f"Data sent to FAKE serial port: {data_to_send}")
            return

        bytes_to_send = bytes(data_to_send)
        self.serial.write(bytes_to_send)
        logger.debug(f"Data sent to serial port: {data_to_send}")

    @_restore_connection
    def read_from_serial_port(self, response_bytes: int) -> bytes:
        """Reads data from the serial port.

        :param response_bytes: The number of bytes to read from the serial port

        :returns: The response from the serial port
        """
        if settings.EMULATE_DEVICES:
            response = bytes([0x00] * response_bytes)
            logger.debug(f"Data received from FAKE serial port: {list(response)}")
            return response

        response = self.serial.read(response_bytes)
        logger.debug(f"Data received from serial port: {list(response)}")
        return response

    def communicate_with_serial_port(self, data_to_send: list[int], response_bytes: int) -> bytes:
        """Communicates with the serial port by sending data and receiving a response.

        :param data_to_send: The data to send to the serial port
        :param response_bytes: The number of bytes to read from the serial port as a response

        :returns: The response from the serial port
        """
        self.write_to_serial_port(data_to_send)
        response = self.read_from_serial_port(response_bytes)
        return response

    def _bytes_to_int(self, bytes_: bytes) -> int:
        """Converts a byte array to an integer.

        :param bytes_: The byte array to convert to an integer

        :returns: The integer representation of the byte array
        """
        return int.from_bytes(bytes_, byteorder="big")

    def _int_to_bytes(self, integer: int, n_bytes: int | None = None) -> list[int]:
        """Converts an integer to a byte array.

        :param integer: The integer to convert to a byte array
        :param n_bytes: The number of bytes to represent the integer. Defaults to None.

        :returns: The byte array representation of the integer
        """
        if n_bytes is None:
            n_bytes = (integer.bit_length() + 7) // 8

        byte_representation = integer.to_bytes(n_bytes, byteorder="big")
        return list(byte_representation)

    def __del__(self):
        """Closes the serial connection when the object is deleted."""
        if settings.EMULATE_DEVICES:
            logger.debug(f"Closing FAKE serial connection with {self.port}")
            return

        logger.debug(f"Closing serial connection with {self.port}")
        self.serial.close()



import threading

import ttkbootstrap as ttk
from bioexperiment_suite.interfaces import Pump, Spectrophotometer
from bioexperiment_suite.tools import get_connected_devices
from ttkbootstrap import constants as c

from .device_widgets import PumpWidget, SpectrophotometerWidget
from .store import Store


class ConnectedDevicesWidget(ttk.Frame):
    FRAME_PADDING = 5
    PADX = 5
    PADY = 5

    def __init__(self, parent: ttk.Frame, store: Store):
        super().__init__(parent, padding=5)

        self.store = store

        self.pumps: list[Pump] = []
        self.spectrophotometers: list[Spectrophotometer] = []

        self.create_widgets()

    def discover_devices(self) -> None:
        for device_widgets in (self.store.pump_widgets, self.store.spectrophotometer_widgets):
            while device_widgets:
                widget = device_widgets.pop()
                widget.device.__del__()
                widget.destroy()

        progress = ttk.Progressbar(self.devices_frame, mode="determinate", bootstyle=c.STRIPED)
        progress.pack(fill=c.X, expand=c.YES, padx=self.PADX, pady=self.PADY)
        self.master.update()

        for class_ in (PumpWidget, SpectrophotometerWidget):
            class_.instance_count = 0  # type: ignore

        def find_connected_devices():
            self.pumps, self.spectrophotometers = get_connected_devices()

        t = threading.Thread(target=find_connected_devices)

        t.start()
        while t.is_alive():
            progress.step(1)
            self.master.update()
            t.join(0.1)

        progress.destroy()

        while self.pumps:
            pump = self.pumps.pop()
            pump_widget = PumpWidget(self.devices_frame, pump)
            self.store.pump_widgets.append(pump_widget)
            pump_widget.pack(side=c.LEFT, fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

        while self.spectrophotometers:
            spec = self.spectrophotometers.pop()
            spec_widget = SpectrophotometerWidget(self.devices_frame, spec)
            self.store.spectrophotometer_widgets.append(spec_widget)
            spec_widget.pack(side=c.LEFT, fill=c.X, expand=c.NO, padx=self.PADX, pady=self.PADY)

    def create_widgets(self) -> None:
        discover_button = ttk.Button(self, text="Discover devices", command=self.discover_devices, bootstyle=c.PRIMARY)
        discover_button.pack(fill=c.NONE, expand=c.NO, padx=self.PADX, pady=self.PADY)

        self.devices_frame = ttk.Frame(self, padding=5)
        self.devices_frame.pack(fill=c.X, expand=c.YES, padx=self.PADX, pady=self.PADY)



from datetime import datetime, timedelta
from typing import Any, Callable

from bioexperiment_suite.loader import logger


class Action:
    """Class to define an action to be executed in an experiment.

    The action can be executed with the `execute` method, which will call the function with the provided arguments.
    The action keeps track of the time it was started and completed.
    """

    def __init__(self, func: Callable, *args: Any, **kwargs: Any) -> None:
        """Initialize the action with the function to be executed and the arguments to be passed to it.

        :param func: The function to be executed
        :param args: The positional arguments to be passed to the function
        :param kwargs: The keyword arguments to be passed to the function
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.start_time: datetime | None = None
        self.end_time: datetime | None = None
        logger.debug(f"Action created: {self.func.__name__} with args: {args} and kwargs: {kwargs}")

    def execute(self) -> None:
        """Execute the action by calling the function with the provided arguments and tracking the start and end time."""
        self.start_time = datetime.now()
        logger.debug(f"Executing action: {self.func.__name__}")
        self.func(*self.args, **self.kwargs)
        self.end_time = datetime.now()
        logger.debug(f"Action completed: {self.func.__name__}")

    def is_completed(self) -> bool:
        """Check if the action has been completed.

        :return: True if the action has been completed, False otherwise
        """
        return self.end_time is not None and self.start_time is not None

    def duration(self) -> timedelta:
        """Get the duration of the action.

        :return: The duration of the action as a timedelta

        :raises ValueError: If the action has not been completed yet
        """
        if not self.is_completed():
            raise ValueError("Action did not complete yet")

        return self.end_time - self.start_time  # type: ignore


class Measurement(Action):
    """Class to define a measurement to be executed in an experiment.

    The measurement is a special type of action that also stores the measured value name and the measurement result.
    """

    def __init__(self, func: Callable, measurement_name: str, *args: Any, **kwargs: Any):
        """Initialize the measurement with the function to be executed, the measured value name, and the arguments.

        :param func: The function to be executed
        :param measurement_name: The name of the measured value
        :param args: The positional arguments to be passed to the function
        :param kwargs: The keyword arguments to be passed to the function
        """
        super().__init__(func, *args, **kwargs)
        self.measurement_name = measurement_name
        self.measured_value = None
        logger.debug(f"Measurement created: {self.func.__name__} with args: {args} and kwargs: {kwargs}")

    def execute(self) -> Any:
        """Execute the measurement by calling the function with the provided arguments and tracking the start and end time.

        :return: The result of the measurement
        """
        assert self.start_time is None, "Measurement already executed"
        self.start_time = datetime.now()
        logger.debug(f"Executing measurement: {self.func.__name__}")
        result = self.func(*self.args, **self.kwargs)
        self.end_time = datetime.now()
        self.measured_value = result
        logger.debug(f"Measurement completed: {self.func.__name__}")
        return result


class WaitAction:
    """Class to define a wait action to be executed in an experiment.

    The wait action is a special type of action that pauses the execution of the experiment for a given amount of time.
    If some action takes time to be executed, the wait time should be not less than the time taken by the action.
    """

    def __init__(self, seconds: float):
        """Initialize the wait action with the number of seconds to wait.

        :param seconds: The number of seconds to wait
        """
        self.wait_time: timedelta = timedelta(seconds=seconds)
        logger.debug(f"Wait action created: {seconds} seconds")



from __future__ import annotations

import inspect
import os
import time
from collections import defaultdict
from datetime import datetime
from threading import Event, Thread
from typing import Any, Callable, get_type_hints

from bioexperiment_suite.loader import logger

from .actions import Action, WaitAction, Measurement
from .collections import Statistic, RelationFunction


class Metric:
    """Class to define a metric to be used during the experiment run to make dynamic decisions"""

    def __init__(self, experiment: Experiment, measurement_name: str, statistic: Statistic | None = Statistic.LAST()):
        """Initialize the metric to be used in the experiment

        :param experiment: The experiment object
        :param measurement_name: The name of the measurement in the experiment to use for the metric calculation
        :param statistic: The statistic to apply to the measurement values
        """
        self.measurements = experiment.measurements
        self.measurement_name = measurement_name
        self.statistic = statistic

    def _measurement_values(self) -> tuple[float]:
        return tuple(zip(*self.measurements[self.measurement_name]))[1]

    def get_value(self) -> int | float:
        return self.statistic(self._measurement_values())


class Condition:
    """Class to define a condition to determine if an action should be executed."""

    def __init__(self, metric: Metric, relation: RelationFunction):
        """
        Initialize the condition for the action to be executed.

        :param metric: The metric to evaluate.
        :param relation: A callable that takes the metric's value and returns a boolean.
        """
        if not isinstance(metric, Metric):
            raise TypeError("Expected a Metric instance for 'metric'.")
        if not callable(relation):
            raise TypeError("Expected a callable for 'relation'.")

        self.metric = metric
        self.relation = relation

    def check_condition(self) -> bool:
        """
        Check if the condition is met.

        :return: True if the condition is met, False otherwise.
        """
        metric_value = self.metric.get_value()
        return self.relation(metric_value)

    @property
    def negation(self) -> Condition:
        """Return a new condition that is the negation of this condition."""
        return Condition(self.metric, lambda x: not self.relation(x))


class ConditionalAction:
    """Class to define action to be executed based on a condition."""

    def __init__(self, action: Action, condition: Condition):
        """Initialize the conditional action with the condition and the action to be executed

        :param condition: The condition to evaluate
        :param action: The action to execute if the condition is met
        """
        self.action = action
        self.condition = condition
        logger.debug(f"Conditional action created: {self.condition} -> {self.action}")

    def get_action(self) -> Action | None:
        """Return the action to be executed if the
        condition is met, otherwise return None."""
        if self.condition.check_condition():
            return self.action
        return None


class Experiment:
    """Class to define an experiment with actions and measurements to be executed in sequence.

    The experiment can be run with the `run` method, which will execute each action in sequence.
    The experiment keeps track of the time each action was executed and the measurements taken.
    """

    def __init__(self, output_dir: os.PathLike | None = None):
        """Initialize the experiment with an empty list of actions and measurements"""
        self.actions: list[Action | WaitAction | ConditionalAction] = []
        self.measurements: dict[str, list[tuple[datetime, Any]]] = defaultdict(list)
        self.current_time: datetime | None = (
            None  # Time to keep track of the experiment progress. Initializes on start.
        )
        self.output_dir = output_dir
        self._thread: Thread | None = None
        self._stop_event = Event()
        logger.debug("Experiment created")

    def create_metric(self, measurement_name: str, statistic: Statistic | None = Statistic.LAST()) -> Metric:
        """Create a metric object to be used in the experiment for making dynamic decisions.

        :param measurement_name: The name of the measurement in the experiment to use for the metric calculation
        :param statistic: The statistic to apply to the measurement values

        :return: The metric object
        """
        return Metric(self, measurement_name, statistic)

    def specify_output_dir(self, output_dir: os.PathLike):
        """Specify the output directory to write measurements to CSV files.

        :param output_dir: The directory to write measurements to
        """
        self.output_dir = output_dir
        logger.debug(f"Output directory specified: {output_dir}")

    def add_action(self, func: Callable, condition: Condition | None = None, *args: Any, **kwargs: Any):
        """Add an action to the experiment.

        The action will be executed in sequence when the experiment is run.

        :param func: The function to be executed
        :param condition: The condition to evaluate before executing the action
        :param args: The positional arguments to be passed to the function
        :param kwargs: The keyword arguments to be passed to the function
        """
        self._validate_types(func, *args, **kwargs)
        action = Action(func, *args, **kwargs)
        if condition is not None:
            self.actions.append(ConditionalAction(action, condition))
            logger.debug(f"Conditional action added to experiment: {func.__name__} with condition: {condition}")
            return

        self.actions.append(action)
        logger.debug(f"Action added to experiment: {func.__name__}")

    def add_measurement(
        self, func: Callable, measurement_name: str, condition: Condition | None = None, *args: Any, **kwargs: Any
    ):
        """Add a measurement to the experiment.

        The measurement will be executed in sequence when the experiment is run, and the result will be stored.

        :param func: The function to be executed
        :param measurement_name: The name of the measured value
        :param condition: The condition to evaluate before executing the action
        :param args: The positional arguments to be passed to the function
        :param kwargs: The keyword arguments to be passed to the function
        """
        self._validate_types(func, *args, **kwargs)
        measurement = Measurement(func, measurement_name, *args, **kwargs)
        if condition is not None:
            self.actions.append(ConditionalAction(measurement, condition))
            logger.debug(f"Conditional measurement added to experiment: {func.__name__} with condition: {condition}")
            return

        self.actions.append(measurement)
        logger.debug(f"Measurement added to experiment: {func.__name__}")

    def add_wait(self, seconds: float, condition: Condition | None = None):
        """Add a wait action to the experiment.

        The wait action will pause the execution of the experiment for the given amount of time.

        :param seconds: The number of seconds to wait
        :param condition: The condition to evaluate before executing the action
        """
        wait_action = WaitAction(seconds)
        if condition is not None:
            self.actions.append(ConditionalAction(wait_action, condition))
            logger.debug(f"Conditional wait added to experiment: {seconds} seconds with condition: {condition}")
            return

        self.actions.append(wait_action)
        logger.debug(f"Wait action added to experiment: {seconds} seconds")

    def _perform_action(self, action: Action | WaitAction | ConditionalAction, step: int) -> bool:
        """Perform the action by executing it or waiting for the specified time.

        :param action: The action to perform
        """
        logger.debug(f"Step {step + 1} from {len(self.actions)}")
        if isinstance(action, Measurement):
            logger.debug(f"Executing measurement: {action.func.__name__}")
            action.execute()
            self.measurements[action.measurement_name].append((datetime.now(), action.measured_value))
            self.write_measurement_to_csv(action.measurement_name)
        elif isinstance(action, Action):
            logger.debug(f"Executing action: {action.func.__name__}")
            action.execute()
        elif isinstance(action, WaitAction):
            wait_until = self.current_time + action.wait_time
            logger.debug(f"Waiting for {action.wait_time.total_seconds()} seconds from {self.current_time}")

            if datetime.now() > wait_until:
                logger.warning(f"Wait time exceeded on step {step + 1} by {datetime.now() - wait_until}")

            while datetime.now() < wait_until:
                if self._stop_event.is_set():
                    logger.debug("Experiment stopped")
                    return False
                time.sleep(0.1)

            self.current_time += action.wait_time

        elif isinstance(action, ConditionalAction):
            action_to_execute = action.get_action()
            if action_to_execute is not None:
                self._perform_action(action_to_execute, step)
            else:
                logger.debug("Condition not met, skipping action")

        else:
            logger.error(f"Unknown action type: {type(action)}")
            raise ValueError(f"Unknown action type: {type(action)}")

        return True

    def _run(self):
        """Run the experiment by executing each action in sequence."""
        self.current_time = datetime.now()
        logger.debug(f"Experiment started. Start time: {self.current_time}")
        for step, action in enumerate(self.actions):
            if not self._perform_action(action, step):
                return

    def start(self, start_in_background: bool = True):
        """Start the experiment by running it in idle mode or in a separate thread depending on the `start_in_background` flag.

        :param start_in_background: If True, start the experiment in a separate thread.
        If False, run the experiment in the current thread. Be careful with this option,
        as it will block the current thread until the experiment is finished!
        """
        if self._thread is not None:
            logger.warning("Experiment is already running")
            return

        self._stop_event.clear()
        if start_in_background:
            self._thread = Thread(target=self._run)
            self._thread.start()
        else:
            self._run()

    def stop(self):
        """Stop the experiment by setting the stop event."""
        if self._thread is None:
            logger.warning("Experiment is not running")
            return
        self._stop_event.set()
        logger.info("Experiment stop signal sent")

    def write_measurement_to_csv(self, measurement_name: str):
        """Write last acquired measurement to a CSV file.

        :param measurement_name: The name of the measurement to write to a CSV file
        """
        if self.output_dir is None:
            return

        if measurement_name not in self.measurements:
            raise ValueError(f"Measurement '{measurement_name}' not found")

        output_file = os.path.join(self.output_dir, f"{measurement_name}.csv")
        with open(output_file, "a") as f:
            timestamp, value = self.measurements[measurement_name][-1]
            f.write(f"{timestamp},{value}\n")

        logger.debug(f"Measurement '{measurement_name}' written to {output_file}")

    def reset_experiment(self):
        """Reset the experiment by clearing the actions, measurements and current time."""
        logger.debug("Experiment reset")
        if self._thread is not None:
            logger.warning("Experiment is running. Stop it before resetting.")
            return

        self.actions.clear()
        self.measurements.clear()
        self.current_time = None

    def _validate_types(self, func: Callable, *args: Any, **kwargs: Any):
        """Validate that the arguments passed to the function are of the correct type.

        :param func: The function to validate the arguments for
        :param args: The positional arguments to validate
        :param kwargs: The keyword arguments to validate

        :raises TypeError: If any argument is not of the expected type according to the type hints of the function
        """
        type_hints = get_type_hints(func)
        sig = inspect.signature(func)
        bound_arguments = sig.bind_partial(*args, **kwargs).arguments

        for name, value in bound_arguments.items():
            expected_type = type_hints.get(name)
            if expected_type and not isinstance(value, expected_type):
                msg = f"Argument '{name}' is expected to be of type {expected_type}, but got {type(value)}"
                raise TypeError(msg)



import os

from dataclasses import dataclass
from dotenv import load_dotenv
from functools import lru_cache


@dataclass
class Settings:
    EMULATE_DEVICES: bool = False


@lru_cache
def get_settings():
    load_dotenv()
    return Settings(EMULATE_DEVICES=os.getenv("EMULATE_DEVICES", "False").lower() == "true")


settings = get_settings()



import json
import sys
from importlib import resources as impresources

from loguru import logger
from munch import DefaultMunch

import bioexperiment_suite

logger.remove(0)
logger.add(
    sys.stderr,
    level="INFO",
)

device_interfaces_file = impresources.files(bioexperiment_suite) / "device_interfaces.json"
with device_interfaces_file.open(encoding="utf8") as file:
    device_interfaces = DefaultMunch.fromDict(json.load(file))
